module ActionController
  module Macros
    module ImageMagick #:nodoc:
      def self.append_features(base) #:nodoc:
        super
        
        base.extend(ClassMethods)
      end

      # This macro can be used to add ImageMagick functionality to a controller.
      # You can use it to access ImageMagick commands with a simple template tag.
      # ImageMagick is a set of image processing tools, that can do nearly everything
      # you'd ever want to do with an image. Resize, rotate, crop and much more.
      #
      # This needs the RMagick library (and ImageMagick, of course) to work.
      #
      # * http://www.imagemagick.org/
      # * http://rmagick.rubyforge.org/
      #
      # Example:
      #
      #   # Controller
      #   class Photo < ApplicationController
      #     imagemagick_for '/var/lib/photos'
      #   end
      #
      #   # View
      #   <%= imagemagick_tag 'presentation.jpg', 'resize(100x50)+flop' %>
      #
      # This example generates an <tt>&lt;img&gt;</tt>-tag in the view, which points to
      # the +imagemagick+ action of the Photo controller. All you need to do to
      # generate this controller, is adding the <tt>imagemagick_for</tt> method to your
      # controller class. The +imagemagick+ action opens the image, sends the
      # commands to RMagick for processing and then sends the result to the browser.
      #
      # This example shows you the image presentation.jpg from the /var/lib/photos
      # directory. It is processed with the commands +resize(100x50)+, which means
      # that ImageMagick will shrink it to fit it in an area of 100x50 pixels, and
      # +flop+, which means that left and right are reversed.
      #
      #
      # == The image directory
      #
      # This extension expects to find all images in one directory, the name of which
      # is specified with the +imagemagick_for+ method. In this directory, you can simply
      # place the image files. The names and extensions are not really important,
      # ImageMagick is smart enough to determine the file type from the contents of the file.
      #
      # In the image directory, you can have subdirectories. Just give the name of the
      # subdirectory to +imagemagick_tag+. For example, if in our directory +/var/lib/photos+
      # there is a subdirectory called rails, we can refer to an image in that subdirectory
      # like this:
      #
      #   # View
      #   <%= imagemagick_tag 'rails/presentation.jpg' %>
      #
      # You can use the ImageMagick extension in multiple controllers, but you might
      # want to specify a different image directory for each of them. (But that's not
      # compulsory.)
      #
      #
      # == Caching
      #
      # If you don't take any precautions, the rendered image will not be saved, and
      # it has to be rendered from scratch for every new request. This can make it a
      # little slow. This extension therefore provides a caching mechanism, which caches
      # the processed images. For the following requests, it simply sends the cached
      # image, instead of re-rendering it.
      #
      # To enable caching, you need a cache directory that is writable by the webserver.
      # If you have that directory, all you have to do is specify it with the +cache+ option
      # of the +imagemagick_for+ method.
      #
      #   # Controller
      #   class Photo < ApplicationController
      #     imagemagick_for '/var/lib/photos', :cache=>'/var/lib/photo-cache'
      #   end
      #
      # If you enable caching, you need a separate cache directory for every controller
      # that uses the +imagemagick_tag+.
      #
      #
      # == Pre-rendering
      #
      # By default, the commands are transferred in the URL of the image, and are executed
      # when the user requests the image. This is reasonably safe, but it does allow your
      # visitors to specify any ImageMagick command they want by simply changing the URL.
      # If you think this is a problem, you can enable the pre-rendering function.
      #
      # Pre-rendering means that the image is rendered earlier in the process: when the
      # +imagemagick_tag+ is displayed by the view. The image is rendered and cached, and the
      # name of that cached image is sent to the browser, which then uses it to request the
      # image. This means that the user does not have the possibility to influence the
      # rendering of the image, as rendering is done from the view. It also means that the
      # initial delay of the rendering of the image is on the web page, instead of on the
      # loading of the images.
      #
      # The pre-rendered images are stored by the caching system, so you'll need to
      # enable caching if you want to use pre-rendering.
      #
      #   # Controller
      #   class Photo < ApplicationController
      #     imagemagick_for '/var/lib/photos', :cache=>'/var/lib/photo-cache', :prerender=>true
      #   end
      # 
      #
      #
      # == Creating recipes
      #
      # Adds a custom 'recipe' to this controller's ImageMagick. With a recipes,
      # you can create custom commands that you can use with the +imagemagick_tag+.
      #
      # Your own recipes can have arguments and you can mix them with normal
      # ImageMagick methods. Such as:
      #
      #   <%= imagemagick_tag 'picture.jpg', 'customthumbnail+resize(100)+myborder(f00)' %>
      #
      # There are multiple ways to create your own recipes: as a String of normal
      # commands, as an inline method (proc), as a method reference (symbol) or as
      # an external class.
      #
      # <strong>If you use the caching option, you should clean the cache if you change
      # existing recipes.</strong> The recipe name will be saved in the cache, so if you
      # only change the recipe contents, the cached image will not be re-rendered.
      #
      #
      # === String
      #
      # The easiest way is <strong>with a String</strong>. If you always render your
      # profile images with <tt>'resize(100)+border(1,1,fff)'</tt>, you can shorten
      # that with this recipe:
      #
      #   # in the controller
      #   class UserController
      #     imagemagick_for '/var/lib/profiles'
      #     imagemagick_recipe :userprofile, 'resize(100)+border(1,1,fff)'
      #   end
      #   
      #   # in the view
      #   <%= imagemagick_tag 'john.jpg', 'userprofile' %>
      # 
      # As you see, you've now created your own 'command'. +userprofile+ will be
      # expanded to <tt>resize(100)+border(1,1,fff)</tt> when rendering.
      # 
      #
      # === Proc
      #
      # A more sophisticated method is by using a Proc. You can use this
      # if you want more control than you can get with the String method. And if you
      # need custom parameters for your recipes, you simply can't use a String.
      #
      # With a proc, adding a recipe works like this:
      #
      #   # in the controller
      #   class UserController
      #     imagemagick_for '/var/lib/profiles'
      #     imagemagick_recipe :square, Proc.new { |image, size| image.resize!(size.to_i, size.to_i) }
      #   end
      #   
      #   # in the view
      #   <%= imagemagick_tag 'john.jpg', 'square(100)' %>
      # 
      # This example renders the image as a 100x100 pixel square. (If you look closely, you'll
      # notice that we're passing the size of 100 pixel as a parameter to the recipe.)
      # Proc-recipes work by accepting an +image+ object and perhaps some parameters. In this case,
      # the only parameter is +size+. The +image+ object is the real <tt>RMagick::Image</tt>-object
      # of the image that is currently being processed. In your proc, you can execute any of the
      # <a href="http://studio.imagemagick.org/RMagick/doc/">RMagick methods</a> you like.
      # 
      # There are two caveats you should keep in mind when using Proc recipes:
      #
      # * Except for the +image+ object, <strong>all arguments passed to the Proc are Strings.</strong>
      #   The incoming parameter string is split by the commas, so 100,100 would give you
      #   two arguments. In your proc the "100" and "100" are +String+s, but most RMagick commands
      #   expect +Integer+s or +Float+s. Remember to convert them with +to_i+ or +to_f+ (as in
      #   the example above).
      # 
      # * <strong>Your proc should return the +image+ object.</strong> Not every RMagick method is
      #   available as an in-place method (which is destructive, with a trailing !). Thus, to make
      #   sure that the changes your recipe makes are saved, you should return the changed image.
      #   (When a destructive method is available, use that, but returning the image object is still
      #   required.)
      #
      # 
      # === Method and Class
      #
      # If you don't like providing the recipe as a Proc, you can also give your recipes the form
      # of methods of the controller, or as a separate class.
      #
      # If using a method, give the method name as a Symbol:
      #
      #   class UserController
      #     imagemagick_for '/var/lib/profiles'
      #     imagemagick_recipe :ownresize, :my_own_resize_method
      #     
      #     private
      #       def my_own_resize_method(image, width, height)
      #         image.resize!(width.to_i, height.to_i)
      #       end
      #   end
      #
      # If using a class, it should have a class method called +execute_recipe_on+.
      # Provide the class to imagemagick_recipe:
      #
      #   class MyResizeRecipe
      #     def self.execute_recipe_on(image, width, height)
      #       image.resize!(width.to_i, height.to_i)
      #     end
      #   end
      #
      #   class UserController
      #     imagemagick_for '/var/lib/profiles'
      #     imagemagick_recipe :ownresize, MyResizeRecipe
      #   end
      #
      # As for the arguments and the expected return, the method and class approaches are
      # exactly the same as with a proc: 
      #
      #
      # == Routes.rb trick
      #
      # +imagemagick_for+ works without any additional routes. It uses urls like:
      #   /photo/imagemagick/presentation.jpg?command=resize(100x100)
      #
      # If you'd like to get rid of the <tt>?command=</tt>, you can add the following routes to
      # your <tt>routes.rb</tt>:
      #     map.connect ':controller/imagemagick/:id', :action=>'imagemagick', :commands=>''
      #     map.connect ':controller/imagemagick/:commands/:id', :action=>'imagemagick'
      #
      # The above url will now be rendered as:
      #   /photo/imagemagick/resize(100x100)/presentation.jpg
      #
      module ClassMethods
        #
        # Enables the ImageMagick extension for this controller. The only thing you
        # really need is the +image_path+ parameter, which points to the directory with
        # images for this controller. If you want to, you can also specify the following +options+:
        #
        # * <tt>:cache</tt>: the directory where cached images are stored, or +nil+ (the default)
        #   if you don't want caching. The directory has to be writable by the web server.
        # * <tt>:prerender</tt>: set this to +true+ if you want pre-rendering. This only works if
        #   you have enabled caching. By default, pre-rendering is off.
        # * <tt>:action_name</tt>: the name of the imagemagick action. The default is +imagemagick+,
        #   you can change it by specifying this option.
        #
        def imagemagick_for(image_path, options = {})
          # strip the trailing slash
          image_path = image_path.sub(/\/$/, '') if image_path =~ /\/$/
          
          require 'RMagick'
          
          configuration = { :image_path=>image_path, :cache => nil, :prerender=>false, :action_name=>:imagemagick }
          configuration.update(options)
          configuration[:image_path] = File.expand_path(configuration[:image_path], RAILS_ROOT)
          configuration[:cache] = File.expand_path(configuration[:cache], RAILS_ROOT) unless configuration[:cache].nil?
          configuration[:action_name] = configuration[:action_name].to_sym
          
          #
          # Every controller-class gets a imagemagick_macro_helper-variable,
          # which contains the configuration options.
          #
          @imagemagick_macro_helper = ImageMagickMacroHelper.new(configuration)
          
          #
          # Makes the class helper available from the controller instance.
          #
          define_method(:imagemagick_macro_helper) do
            return self.class.instance_variable_get(:@imagemagick_macro_helper)
          end
          protected :imagemagick_macro_helper
          
          #
          # Shortcut to the +url_for_imagemagick+ method of the helper.
          # Gives the url to the +imagemagick+ action for this filename, with these commands.
          #
          define_method(:url_for_imagemagick) do |filename, p_commands|
            url_for(imagemagick_macro_helper.url_options_for_imagemagick(filename, p_commands))
          end
          
          #
          # Where it all takes place: the +imagemagick+ action of the controller.
          # Reads the filename and the commands from the request and sends an image.
          #
          define_method(configuration[:action_name]) do
            commands = ImageMagickMacroHelper::MagickCommandList.from_anything(@params["commands"])
            filename = @params["id"]
            
            # set the controller object in the helper,
            # needed for the helper to execute methods of the controller
            imagemagick_macro_helper.controller = self
            
            # check the filename: it may contain directory names, but it cannot
            # go to a higher directory than configuration[:image_path]
            pathname = File.expand_path(filename, "#{configuration[:image_path]}")
            unless pathname =~ /^#{configuration[:image_path]}\//
              render :text=>"Image not found.", :status=>404
              return
            end
            
            # is prerendering enabled?
            if imagemagick_macro_helper.prerender?
              # yes, we only accept requests for images that are already in the cache
              cache_path = "#{configuration[:cache]}/" + imagemagick_macro_helper.cache_filename(filename, commands)
              
              if File.exists?(cache_path)
                # determine the mime type and send the file
                @response.headers["Content-type"] = Magick::Image.read(cache_path).first.mime_type
                render :file=>cache_path
              else
                render :text=>"Image not found.", :status=>404
              end
              
            else
              # no, we might have to render the image
              if filename =~ /^\.*$/ || !File.readable?(pathname)
                render :text=>"Image not found.", :status=>404
              else
                # the helper finds and/or renders the image
                picture = imagemagick_macro_helper.render(filename, commands)
                
                if picture.is_a?(String)
                  # the image is already cached, render from file
                  @response.headers["Content-type"] = Magick::Image.read(picture).first.mime_type
                  render :file=>picture
                  
                else
                  # the image was not yet cached, render from memory
                  @response.headers["Content-type"] = picture.mime_type
                  render :text=>picture.to_blob
                end
              end
            end
          end
        end
        
        #
        # Adds a custom recipe with the given +name+. +recipe+ can be a +String+,
        # a +Proc+, a +Class+ or a +Symbol+.
        #
        def imagemagick_recipe(name, recipe)
          if @imagemagick_macro_helper.nil?
            raise(ActionControllerError, "Use imagemagick_for before you use imagemagick_recipe.")
          end
          
          @imagemagick_macro_helper.add_recipe(name, recipe)
        end
        
      end
      
      #
      # The class that contains all methods of the ImageMagick extension.
      # Every controller with imagemagick_for gets an instance of this, filled
      # with its own configuration options.
      #
      class ImageMagickMacroHelper #:nodoc: all
        def initialize(configuration)
          @configuration = configuration
          @configuration[:recipes] = Hash.new
          @configuration[:recipes].merge!(standard_recipes)
        end
        
        def caching?
          return !@configuration[:cache].nil?
        end
        
        def prerender?
          # only if caching is available
          return caching? && @configuration[:prerender]
        end
        
        def controller=(controller)
          @configuration[:controller] = controller
        end
        
        def add_recipe(name, action)
          @configuration[:recipes][name.to_sym] = action
        end
        
        #
        # Returns a Hash of url_for options that render the given filename
        # with the given ImageMagick commands.
        #
        def url_options_for_imagemagick(filename, p_commands)
          commands = MagickCommandList.from_anything(p_commands)
          path_options = { :action=>@configuration[:action_name].to_s }
          
          if prerender?
            path_options[:id] = cache_filename(filename, commands)
            render(filename, commands)
          else
            render(filename, commands)
            path_options[:id] = filename
            path_options[:commands] = commands.to_s unless commands.commands.empty?
          end
          
          return path_options
        end
        
        #
        # Returns the filename the cached image will have if it is renderd
        # with the given commands.
        #
        def cache_filename(filename, commands)
          commands = MagickCommandList.from_anything(commands)
          return commands.to_cache_s + filename.gsub(/\//, '.')
        end
        
        #
        # Renders the image with the given commands.
        # The result is saved in the cache, if that is enabled
        #
        # The function returns either an Magick::Image-object of the image or
        # a String, which points to the already-cached image.
        #
        def render(filename, commands)
          pathname = @configuration[:image_path] + "/"+ filename
          
          cache_path = nil
          if caching?
            cache_path = @configuration[:cache] + "/" + cache_filename(filename, commands)
          end
          
          if (!caching? || !File.exists?(cache_path) || File.mtime(cache_path) < File.mtime(pathname))
            pic = Magick::Image.read(pathname).first
            
            pic = commands.execute_on(pic, @configuration)
            
            pic.strip!
            
            if caching?
              pic.write(cache_path)
            end
            
            return pic
          else
            return cache_path
          end
        end
        
        #
        # Adds recipes for the standard RMagick commands.
        #
        def standard_recipes
          r = Hash.new
          
          # standard RMagick commands
          r[:blur_image] = Proc.new { |image, radius, sigma| image.blur_image(radius.to_f, sigma.to_f) }
          r[:border] = Proc.new do |image, width, height, color|
            color = "#" + color unless color[0,1]=="#"
            image.border!(width.to_i, height.to_i, color)
          end
          r[:colorize] = Proc.new do |image, red_pct, green_pct, blue_pct, fill|
            fill = "#" + fill unless fill[0,1]=="#"
            image.colorize(red_pct.to_f, green_pct.to_f, blue_pct.to_f, fill)
          end
          r[:crop] = Proc.new { |image, x, y, width, height| image.crop!(x.to_i, y.to_i, width.to_i, height.to_i) }
          r[:equalize] = Proc.new { |image| image.equalize }
          r[:flip] = Proc.new { |image| image.flip! }
          r[:flop] = Proc.new { |image| image.flop! }
          r[:geometry] = Proc.new { |image, data| image.change_geometry!(data) { |cols, rows, img| img.resize!(cols, rows) } }
          r[:implode] = Proc.new { |image, amount| image.implode(amount.to_f) }
          r[:level] = Proc.new { |image, black_point, mid_point| image.level(black_point.to_f, mid_point.to_f) }
          r[:normalize] = Proc.new { |image| image.normalize }
          r[:oilpaint] = Proc.new { |image, amount| image.oil_paint(amount.to_f) }
          r[:opaque] = Proc.new do |image, target, fill|
            target = "#" + target unless target[0,1]=="#"
            fill = "#" + fill unless fill[0,1]=="#"
            image.opaque(target, fill)
          end
          r[:posterize] = Proc.new { |image, levels| image.posterize(levels.to_f) }
          r[:rotate] = Proc.new { |image, amount| image.rotate!(amount.to_i) }
          r[:sharpen] = Proc.new { |image, radius, sigma| image.sharpen(radius.to_f, sigma.to_f) }
          r[:trim] = Proc.new { |image| image.trim! }
          
          # aliases
          r[:resize] = r[:geometry]
          r[:s] = r[:geometry]
          r[:blur] = r[:blur_image]
          
          
          
          # This method is not available from imagemagick, but
          # it is useful: it will always give you an image of
          # exactly the required size. part(100x100) gives you
          # a 100x100 pixel crop from the center of the image.
          #
          # For example: the image is   ********
          #                             ********
          #                             ********
          # 
          # And you want this:          **
          #                             **
          # It will then first resize the image: ****
          #                                      ****
          # And then crops it:         *[**]*
          #                            *[**]*
          #
          # If the source image was this:
          #        **                    --
          #        **  it would crop:    **
          #        **                    **
          #        **                    --
          #
          # The command also scales up, if the image is too small
          # to cover the whole part.
          r[:part] = Proc.new do |image, size|
            target_w = size.split(/x/)[0].to_i
            target_h = size.split(/x/)[1].to_i
            orig_w = image.columns
            orig_h = image.rows
            
            if (orig_w.to_f/orig_h) > (target_w.to_f/target_h)
              # scale to correct height
              image.change_geometry!("x"+target_h.to_s) { |cols, rows, img| img.resize!(cols, rows) }
              
              # remove left and right
              image.crop!((image.columns - target_w) / 2, 0, target_w, target_h)
              
            else
              # scale to correct width
              image.change_geometry!(target_w.to_s+"x") { |cols, rows, img| img.resize!(cols, rows) }
              
              # remove top and bottom
              image.crop!(0, (image.height - target_h) / 2, target_w, target_h)
            end
            
            image
          end
          
          return r
        end
        
        #
        # Contains a list of ImageMagick commands (of the type MagickCommand).
        #
        class MagickCommandList
          attr_reader :commands
          
          def initialize
            @commands = Array.new
          end
          
          #
          # Returns a String of the commands for use in a url.
          # (The commands are chained with a +.)
          #    => resize(100x1)+chop
          #
          def to_s
            command_string = @commands.join("+")
          end
          
          #
          # Returns a String of the commands for use in the cache-filename.
          # (All strange characters are removed.)
          #    => resize.100x1.chop.
          #
          def to_cache_s
            if !commands.empty?
              command_string = to_s + "."
              command_string.gsub!(/[+,]/, '.')
              command_string.gsub!(/[^a-zA-Z0-9.]/, '')
            else
              command_string = ""
            end
            return command_string
          end
          
          #
          # Executes the commands on this list on the picture.
          # picture should be an Magick::Image-object.
          #
          def execute_on(picture, configuration)
            @commands.each { |cmd| picture = cmd.execute_on(picture, configuration) }
            return picture
          end
          
          #
          # Parses a String of commands and returns the MagickCommandList.
          # String is of the type generated by MagickCommandList::to_s
          #     resize(100x1)+chop
          #
          def self.from_s(command_string)
            list = MagickCommandList.new
            (command_string+"+").scan(/(([^(+]+)(\([^)]+\))?)*/) do |match|
              list.commands << MagickCommand.from_s(match[0]) unless match[0].nil?
            end
            return list
          end
          
          #
          # Returns the MagickCommandList for p_commands.
          # p_commands can be:
          #
          # * a +MagickCommandList+: it will be returned
          # * an +Array+: a list of commands. Commands can be Strings and
          #      arrays: [command, args]:
          #         [ "resize(1x100)", "resize", ["resize", 100] ]
          # * a +String+: will be parsed by MagickCommandList.from_s
          # * anything else: an empty MagickCommandList will be returned
          #
          def self.from_anything(p_commands)
            if p_commands.is_a?(MagickCommandList)
              commands = p_commands
            elsif p_commands.is_a?(Array)
              commands = MagickCommandList.new
              p_commands.each { |cmd| commands.commands << MagickCommand.new(cmd[0].to_s, cmd[1]) }
            elsif p_commands.is_a?(String)
              commands = MagickCommandList.from_s(p_commands)
            else
              commands = MagickCommandList.new
            end
            
            return commands
          end
        end
        
        #
        # A command for in a MagickCommandList.
        # (A command is a command name, and possibly a data string.)
        #
        class MagickCommand
          attr_accessor :command, :data
          
          def initialize(command, data = nil)
            @command = command.to_sym
            @data = data
          end
          
          #
          # Converts the command to a String version for use in the image url.
          #    "command"  or  "command(args)"
          #
          def to_s
            command_string = @command.to_s
            command_string += "(" + MagickCommand.encode(data) + ")" unless data.nil?
            return command_string
          end
          
          #
          # Executes this command on the picture and returns the result.
          #
          def execute_on(picture, configuration)
            if !configuration[:recipes].nil?
              if configuration[:recipes].has_key?(@command.to_sym)
                recipe = configuration[:recipes][@command.to_sym]
                
                # recipes may return Strings. continue solving those
                # Strings until the result is an image, at which point
                # the rendering is finished
                until recipe.is_a?(Magick::Image)
                  # reorganize: Symbol points to a method on the controller
                  #             Class should have a method execute_recipe_on
                  if recipe.is_a?(Symbol)
                    recipe = configuration[:controller].method(recipe)
                  elsif recipe.respond_to?(:execute_recipe_on)
                    recipe = recipe.method(:execute_recipe_on)
                  end
                  
                  if recipe.is_a?(Proc) || recipe.is_a?(Method)
                    recipe = recipe.call(picture, *@data.to_s.split(/,/))
                  elsif recipe.is_a?(String)
                    # a string with other imagemagick commands
                    recipe = MagickCommandList.from_s(recipe).execute_on(picture, configuration)
                  end
                end
                
                picture = recipe
              end
            end
            picture
          end
          
          #
          # Reads a command string (of the MagickCommand::to_s format) and returns
          # a MagickCommand.
          #   resize(100x10)  becomes @command = "resize", @data = "100x10"
          #
          def self.from_s(command)
            return MagickCommand.new(command[/^[^(]+/], command[/\((.+)\)$/, 1])
          end
          
          #
          # Encodes the data string for use in the url command string.
          #
          def self.encode(data)
            # in the command string, ( and ) are used to separate
            # the argument from the command: <tt>resize(100x10)</tt>
            return data.to_s.tr("()", "[]")
          end
          
          #
          # Decodes the data string from the url command string.
          #
          def self.decode(data)
            return data.to_s.tr("[]", "()")
          end
        end
      end
    end
  end
end