require File.dirname(__FILE__) + '/../abstract_unit'

begin
  require "RMagick"
  
  module ActionController::Macros::ImageMagick
    RAILS_ROOT = ""
  end
  
  class SquareTestRecipe
    def self.execute_recipe_on(image, size)
      # no !, to test the return-image process
      image.resize(size.to_i, size.to_i)
    end
  end
  
  class ImageMagickTestController < ActionController::Base
    imagemagick_for File.expand_path(File.dirname(__FILE__) + "/../fixtures/imagemagick"),
                    :cache=>"/tmp/imagemagick_cache"
    
    imagemagick_recipe :string_ten_pixels, "resize(10!x10!)"
    imagemagick_recipe :proc_one_pixel, Proc.new {|image| image.resize!(1,1) }
    imagemagick_recipe :method_square, :method_square
    imagemagick_recipe :class_square, SquareTestRecipe
    
   private
    def method_square(image, size)
      image.resize!(size.to_i, size.to_i)
    end
  end
  
  class ImageMagickSubdirTestController < ActionController::Base
    imagemagick_for File.expand_path(File.dirname(__FILE__) + "/../fixtures")
  end
  
  class ImageMagickCustomNameTestController < ActionController::Base
    imagemagick_for File.expand_path(File.dirname(__FILE__) + "/../fixtures/imagemagick"),
                    :action_name=>"customaction"
  end
  
  class ImageMagickTest < Test::Unit::TestCase
    def setup
      @controller = ImageMagickTestController.new
  
      # enable a logger so that (e.g.) the benchmarking stuff runs, so we can get
      # a more accurate simulation of what happens in "real life".
      @controller.logger = Logger.new(nil)
  
      @request    = ActionController::TestRequest.new
      @response   = ActionController::TestResponse.new
      
      FileUtils.mkdir("/tmp/imagemagick_cache")
    end
    
    def teardown
      FileUtils.rm_rf("/tmp/imagemagick_cache")
    end
  
    def test_no_resize
      get :imagemagick, :id=>"presentation.jpg"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_invalid_filename
      get :imagemagick, :id=>"notfound.jpg"
      assert_response :missing
    end
    
    def test_resize_blur
      get :imagemagick, :id=>"presentation.jpg", :commands=>"s(100x100)+blur(0,1)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 100, pic.columns
    end
    
    def test_crop_colorize
      get :imagemagick, :id=>"presentation.jpg", :commands=>"crop(10,20,40,50)+colorize(0.3,0.3,0.3,dddddd)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 40, pic.columns
      assert_equal 50, pic.rows
    end
    
    def test_equalize_flip_flop
      get :imagemagick, :id=>"presentation.jpg", :commands=>"equalize+flip+flop"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_implode_level_normalize
      get :imagemagick, :id=>"presentation.jpg", :commands=>"implode(0.8)+level(0,1.50)+normalize"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_oilpaint_opaque_posterize
      get :imagemagick, :id=>"presentation.jpg", :commands=>"oilpaint(3.0)+opaque(fff,f00)+posterize(4)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_sharpen_rotate_border
      get :imagemagick, :id=>"presentation.jpg", :commands=>"sharpen(0.3,0.1)+border(2,2,f00)+rotate(-45)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_trim
      get :imagemagick, :id=>"presentation.jpg", :commands=>"trim"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_part
      get :imagemagick, :id=>"presentation.jpg", :commands=>"part(100x100)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 100, pic.columns
      assert_equal 100, pic.rows
    end
    
    def test_cache
      get :imagemagick, :id=>"presentation.jpg", :commands=>"equalize+flip+flop"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      assert File.exists?("/tmp/imagemagick_cache/equalize.flip.flop.presentation.jpg")
      
      get :imagemagick, :id=>"presentation.jpg", :commands=>"equalize+flip+flop"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
    
    def test_subdirectory
      # tests image-filename-with-path
      @controller = ImageMagickSubdirTestController.new
      
      get :imagemagick, :id=>"imagemagick/presentation.jpg"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      get :imagemagick, :id=>"/imagemagick/presentation.jpg"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      get :imagemagick, :id=>"../../imagemagick/presentation.jpg"
      assert_response :missing
    end
    
    def test_recipe_string
      get :imagemagick, :id=>"presentation.jpg", :commands=>"string_ten_pixels"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 10, pic.columns
      assert_equal 10, pic.rows
    end
    
    def test_recipe_proc
      get :imagemagick, :id=>"presentation.jpg", :commands=>"proc_one_pixel"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 1, pic.columns
      assert_equal 1, pic.rows
    end
    
    def test_recipe_method
      get :imagemagick, :id=>"presentation.jpg", :commands=>"method_square(8)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 8, pic.columns
      assert_equal 8, pic.rows
    end
    
    def test_recipe_class
      get :imagemagick, :id=>"presentation.jpg", :commands=>"class_square(16)"
      assert_equal "image/jpeg", @response.headers["Content-type"]
      
      pic = Magick::Image.from_blob(@response.body)[0]
      assert_equal 16, pic.columns
      assert_equal 16, pic.rows
    end
    
    def test_custom_action_name
      @controller = ImageMagickCustomNameTestController.new
      get :customaction, :id=>"presentation.jpg"
      assert_equal "image/jpeg", @response.headers["Content-type"]
    end
  end
  
rescue LoadError
  # RMagick is not installed
end
